import os

import click

from komo import printing
from komo.agent.core import download_workdir


@click.command("download-workdir")
@click.option("--upload-id", type=str, required=True)
def cmd_download_workdir(
    upload_id: str,
):
    download_workdir(
        upload_id,
        os.getcwd(),
    )
    printing.success(f"Workdir {upload_id} successfully downloaded")



import os
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NonNegativeInt,
    PositiveInt,
    ValidationError,
    field_validator,
    model_validator,
)
from pydantic_core import PydanticCustomError


class ClientException(Exception):
    def __init__(self, msg):
        self.msg = msg


class SSHInfoNotFound(ClientException):
    def __init__(self):
        super().__init__("SSH info not found")


class Cloud(str, Enum):
    AWS: str = "aws"
    LAMBDA_LABS: str = "lambda"
    GCP: str = "gcp"
    RUNPOD: str = "runpod"
    KUBERNETES: str = "kubernetes"
    AZURE: str = "azure"
    KOMODO: str = "komodo"


class AcceleratorArgsConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    runtime_version: Optional[str] = Field(default=None)
    tpu_name: Optional[str] = Field(default=None)
    tpu_vm: Optional[bool] = Field(default=None)


class SingleResourcesConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    cloud: Optional[Cloud] = Field(default=None)
    region: Optional[str] = Field(default=None)
    zone: Optional[str] = Field(default=None)
    cpus: Optional[Union[str, PositiveInt]] = Field(default=None)
    memory: Optional[Union[str, PositiveInt]] = Field(default=None)
    accelerators: Optional[Union[str, Dict[str, Optional[PositiveInt]], List[str]]] = (
        Field(default=None)
    )
    instance_type: Optional[str] = Field(default=None)
    use_spot: Optional[bool] = Field(default=None)
    disk_size: Optional[PositiveInt] = Field(default=None)
    disk_tier: Optional[str] = Field(default=None)
    ports: Optional[List[str]] = Field(default=None)
    accelerator_args: Optional[AcceleratorArgsConfig] = Field(default=None)
    image_id: Optional[str] = Field(default=None)

    @field_validator("ports", mode="before")
    @classmethod
    def validate_ports(cls, ports: Any) -> Any:
        if isinstance(ports, str):
            return [ports]

        if isinstance(ports, int):
            return [str(ports)]

        if isinstance(ports, list) and all(
            [isinstance(p, int) or isinstance(p, str) for p in ports]
        ):
            return list(map(str, ports))

        return ports


class ResourcesConfig(SingleResourcesConfig):
    any_of: Optional[List[SingleResourcesConfig]] = Field(default=None)
    ordered: Optional[List[SingleResourcesConfig]] = Field(default=None)


class FileMountConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    source: str
    mode: Literal["MOUNT", "COPY"]


class _TaskConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    # common to all machines/jobs/services
    workdir: Optional[str] = Field(default=None)
    resources: Optional[ResourcesConfig] = Field(default=None)
    envs: Optional[Dict[str, Union[str, int, float]]] = Field(default=None)
    file_mounts: Optional[Dict[str, FileMountConfig]] = Field(default=None)
    setup: Optional[str] = Field(default=None)

    @field_validator("envs", mode="before")
    @classmethod
    def validate_env_var_names(
        cls, envs: Optional[Dict[str, Union[str, int, float]]]
    ) -> Optional[Dict[str, Union[str, int, float]]]:
        if not envs:
            return envs

        for var_name in envs.keys():
            match = re.search("^[a-zA-Z_][a-zA-Z0-9_]*$", var_name)
            if match is None:
                raise PydanticCustomError(
                    "env_regex_mismatch",
                    f"{var_name} is not a valid environment variable name",
                )

        return envs

    @staticmethod
    def _update_config(config: dict, **overrides) -> dict:
        updated_config = config.copy()
        for k, v in overrides.items():
            keys = k.split("/")
            curr_obj = updated_config
            for key in keys[:-1]:
                if key not in curr_obj:
                    curr_obj[key] = {}
                elif not isinstance(curr_obj[key], dict):
                    raise ClientException(
                        f"Cannot set property {k} in config because it already has a"
                        " non-dict value"
                    )

                curr_obj = curr_obj[key]

            curr_obj[keys[-1]] = v

        return updated_config

    @classmethod
    def _from_dict(cls, config: dict, **overrides):
        updated_config = cls._update_config(config, **overrides)
        try:
            config = cls(**updated_config)
        except ValidationError as e:
            error_messages = []
            for error in e.errors():
                loc = error["loc"]
                loc = "/".join(map(str, loc))
                msg = error["msg"]

                error_message = f"'{loc}': {msg}"
                error_messages.append(error_message)

            raise ClientException(
                "\033[1mConfig Error\033[22m\n" + "\n".join(error_messages)
            )

        return config


class MachineConfig(_TaskConfig):
    notebook: Optional[bool] = Field(default=None)

    @classmethod
    def from_yaml(cls, config_file: str, **overrides) -> "MachineConfig":
        if not os.path.isfile(config_file):
            raise ClientException(f"{config_file} does not exist")

        with open(config_file, "r") as f:
            config = yaml.load(f, yaml.FullLoader)

        # ignore job/service config fields
        if "run" in config:
            config.pop("run")
        if "service" in config:
            config.pop("service")

        machine_config = super()._from_dict(config, **overrides)
        return machine_config


class JobConfig(_TaskConfig):
    workdir: Optional[str] = Field(default=None)
    num_nodes: Optional[PositiveInt] = Field(default=None)
    run: str

    @classmethod
    def from_yaml(cls, config_file: str, **overrides) -> "JobConfig":
        if not os.path.isfile(config_file):
            raise ClientException(f"{config_file} does not exist")

        with open(config_file, "r") as f:
            config = yaml.load(f, yaml.FullLoader)

        # ignore machine/service config fields
        if "notebook" in config:
            config.pop("notebook")
        if "service" in config:
            config.pop("service")

        job_config = super()._from_dict(config, **overrides)
        return job_config


class ServiceConfigReadinessProbeSection(BaseModel):
    model_config = ConfigDict(extra="forbid")

    path: str
    post_data: Optional[dict] = Field(default=None)
    initial_delay_seconds: Optional[PositiveInt] = Field(default=None)

    @model_validator(mode="after")
    def validate_readiness_probe(self):
        if not self.path.startswith("/"):
            raise PydanticCustomError(
                "readiness_probe_path_error", "readiness probe path must start with /"
            )
        return self


class ServiceConfigReplicaPolicySection(BaseModel):
    model_config = ConfigDict(extra="forbid")

    min_replicas: NonNegativeInt
    max_replicas: Optional[PositiveInt] = Field(default=None)
    target_qps_per_replica: Optional[PositiveInt] = Field(default=None)
    upscale_delay_seconds: Optional[PositiveInt] = Field(default=None)
    downscale_delay_seconds: Optional[PositiveInt] = Field(default=None)

    @model_validator(mode="after")
    def validate_policy(self):
        if self.max_replicas is not None and self.max_replicas < self.min_replicas:
            raise PydanticCustomError(
                "min_max_replica_error",
                "max_replicas must be greater than min_replicas",
            )

        if (
            self.max_replicas is not None
            and self.max_replicas > self.min_replicas
            and self.target_qps_per_replica is None
        ):
            raise PydanticCustomError(
                "target_qps_per_replica_error",
                "target_qps_per_replica must be provided when max_replicas >"
                " min_replicas",
            )

        return self


class ServiceConfigServiceSection(BaseModel):
    model_config = ConfigDict(extra="forbid")

    readiness_probe: ServiceConfigReadinessProbeSection
    replica_policy: ServiceConfigReplicaPolicySection


class ServiceConfig(JobConfig):
    service: ServiceConfigServiceSection

    @classmethod
    def from_yaml(cls, config_file: str, **overrides) -> "ServiceConfig":
        if not os.path.isfile(config_file):
            raise ClientException(f"{config_file} does not exist")

        with open(config_file, "r") as f:
            config = yaml.load(f, yaml.FullLoader)

        # ignore machine config fields
        if "notebook" in config:
            config.pop("notebook")

        service_config = super()._from_dict(config, **overrides)

        return service_config


class JobStatus(Enum):
    PENDING = "pending"
    INITIALIZING = "initializing"
    RUNNING_SETUP = "running_setup"
    RUNNING = "running"
    SHUTTING_DOWN = "shutting_down"
    FINISHED = "finished"
    CANCELLING = "cancelling"
    CANCELLED = "cancelled"
    ERROR = "error"
    UNKNOWN = "unknown"
    NOT_FOUND = "not found"
    UNAUTHORIZED = "unauthorized"
    UNREACHABLE = "unreachable"

    @classmethod
    def executing_statuses(cls):
        return [cls.RUNNING_SETUP, cls.RUNNING]


class MachineStatus(Enum):
    PENDING = "pending"
    INITIALIZING = "initializing"
    RUNNING_SETUP = "running_setup"
    RUNNING = "running"
    STOPPING = "stopping"
    STOPPED = "stopped"
    TERMINATING = "terminating"
    TERINATED = "terminated"
    ERROR = "error"
    UNKNOWN = "unknown"
    NOT_FOUND = "not found"
    UNAUTHORIZED = "unauthorized"
    UNREACHABLE = "unreachable"

    @classmethod
    def executing_statuses(cls):
        return [cls.RUNNING_SETUP, cls.RUNNING]


class ServiceStatus(Enum):
    CONTROLLER_INIT = "CONTROLLER_INIT"
    REPLICA_INIT = "REPLICA_INIT"
    CONTROLLER_FAILED = "CONTROLLER_FAILED"
    READY = "READY"
    SHUTTING_DOWN = "SHUTTING_DOWN"
    FAILED = "FAILED"
    FAILED_CLEANUP = "FAILED_CLEANUP"
    NO_REPLICA = "NO_REPLICA"
    DELETED = "DELETED"
    UNAUTHORIZED = "unauthorized"
    NOT_FOUND = "not found"
    UNKNOWN = "unknown"


class ReplicaStatus(Enum):
    PENDING = "PENDING"
    PROVISIONING = "PROVISIONING"
    STARTING = "STARTING"
    READY = "READY"
    NOT_READY = "NOT_READY"
    SHUTTING_DOWN = "SHUTTING_DOWN"
    FAILED = "FAILED"
    FAILED_INITIAL_DELAY = "FAILED_INITIAL_DELAY"
    FAILED_PROBING = "FAILED_PROBING"
    FAILED_PROVISION = "FAILED_PROVISION"
    FAILED_CLEANUP = "FAILED_CLEANUP"
    PREEMPTED = "PREEMPTED"
    UNKNOWN = "UNKNOWN"


@dataclass
class Job:
    id: str
    status: JobStatus
    status_message: str
    cloud: Optional[Cloud]
    region: Optional[str]
    zone: Optional[str]
    instance_type: Optional[str]
    accelerators: Optional[str]
    ports: Optional[List[dict]]
    disk_size: Optional[int]
    spot: Optional[bool]
    ssh_info: Optional[List[dict]]

    name: str
    num_nodes: int
    requested_resources: dict
    envs: dict
    file_mounts: dict
    setup: str
    run: str

    created_timestamp: int
    started_timestamp: Optional[int]
    updated_timestamp: int
    finished_timestamp: Optional[int]

    @classmethod
    def from_dict(cls, d):
        d["status"] = JobStatus(d["status"])
        if d.get("cloud", None):
            d["cloud"] = Cloud(d["cloud"])

        job = Job(**d)
        return job


@dataclass
class Machine:
    id: str
    status: MachineStatus
    status_message: str
    cloud: Optional[Cloud]
    region: Optional[str]
    zone: Optional[str]
    instance_type: Optional[str]
    accelerators: Optional[str]
    ports: Optional[List[dict]]
    disk_size: Optional[int]
    spot: Optional[bool]
    ssh_info: Optional[dict]

    name: str
    requested_resources: dict
    envs: dict
    file_mounts: dict
    setup: str
    notebook_token: Optional[str]
    notebook_url: Optional[str]

    created_timestamp: int
    started_timestamp: Optional[int]
    updated_timestamp: int
    terminated_timestamp: Optional[int]

    @classmethod
    def from_dict(cls, d):
        d["status"] = MachineStatus(d["status"])
        if d.get("cloud", None):
            d["cloud"] = Cloud(d["cloud"])

        machine = Machine(**d)
        return machine


@dataclass
class ReadinessProbeSection:
    path: str
    initial_delay_seconds: int
    post_data: Optional[dict] = None


@dataclass
class ReplicaPolicySection:
    min_replicas: int
    max_replicas: int
    upscale_delay_seconds: int
    downscale_delay_seconds: int
    target_qps_per_replica: Optional[int] = None


@dataclass
class ServiceSection:
    readiness_probe: ReadinessProbeSection
    replica_policy: ReplicaPolicySection

    @classmethod
    def from_dict(cls, d):
        d["readiness_probe"] = ReadinessProbeSection(**d["readiness_probe"])
        d["replica_policy"] = ReplicaPolicySection(**d["replica_policy"])

        service_section = ServiceSection(**d)
        return service_section


@dataclass
class Service:
    id: str
    status: ServiceStatus
    status_message: str

    name: str
    num_nodes: int
    requested_resources: dict
    envs: dict
    file_mounts: dict
    setup: str
    run: str
    service: ServiceSection

    uptime: int
    active_versions: List[int]

    created_timestamp: int
    updated_timestamp: int

    url: Optional[str] = None

    @classmethod
    def from_dict(cls, d):
        d["status"] = ServiceStatus(d["status"])
        d["service"] = ServiceSection.from_dict(d["service"])

        service = Service(**d)
        return service


@dataclass
class ServiceReplica:
    service_id: str
    replica_id: int
    version: int
    status: Optional[ReplicaStatus]

    cloud: Optional[Cloud]
    region: Optional[str]
    zone: Optional[str]
    instance_type: Optional[str]
    accelerators: Optional[str]
    ports: Optional[List[dict]]
    disk_size: Optional[int]
    spot: Optional[bool]
    ssh_info: Optional[List[dict]]

    created_timestamp: int
    updated_timestamp: int

    @classmethod
    def from_dict(cls, d):
        if d.get("status", None):
            d["status"] = ReplicaStatus(d["status"])
        else:
            d["status"] = None

        if d.get("cloud", None):
            d["cloud"] = Cloud(d["cloud"])

        replica = ServiceReplica(**d)
        return replica



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_job, ssh_job
from komo.types import JobStatus


@click.command("ssh")
@click.argument(
    "job_id",
    type=str,
)
@click.option("--node-index", "-i", type=int, default=0)
@handle_errors
def cmd_ssh(job_id: str, node_index: int):
    ssh_job(job_id, node_index)



import json
import os
import shutil
import tempfile
import time
from typing import Dict, List, Optional, Union

import jwt
import requests
from requests.exceptions import ConnectionError, ConnectTimeout, RequestException
from tenacity import (
    RetryError,
    retry,
    retry_if_exception_type,
    retry_if_result,
    stop_after_attempt,
    wait_exponential,
)

from komo import printing
from komo.cli.utils import (
    _bytes_to_gigabytes,
    _bytes_to_megabytes,
    _get_path_size_and_file_count,
)
from komo.types import (
    ClientException,
    Cloud,
    Job,
    JobConfig,
    JobStatus,
    Machine,
    MachineConfig,
    MachineStatus,
    Service,
    ServiceConfig,
    ServiceReplica,
)
import sentry_sdk
from sentry_sdk import set_user

KOMODO_API_URL = os.environ.get("KOMODO_API_URL", "https://api.komodoai.dev")
KOMODO_JWT_TOKEN_FILE_PATH = os.path.expanduser("~/.komo/jwt-token")
MAX_WORKDIR_SIZE_BYTES = 1073741824  # 1 GB


def _is_retryable_exception(exception):
    return (
        isinstance(exception, ConnectionError)
        or isinstance(exception, ConnectTimeout)
        or isinstance(exception, RequestException)
    )


def _is_retryable_response(response):
    return response.status_code > 500


@retry(
    retry=(
        retry_if_exception_type((ConnectionError, ConnectTimeout, RequestException))
        | retry_if_result(_is_retryable_response)
    ),
    stop=stop_after_attempt(6),
    # Wait exponentially between retries, starting at 1 second and capping at 20 seconds (1,2,4,8,16,20)
    wait=wait_exponential(multiplier=1, min=1, max=20),
)
def _make_request_and_raise_for_status(method, url, headers, files, data):
    try:
        response = requests.request(
            method,
            url,
            headers=headers,
            files=files,
            data=data,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code <= 500:
            text = e.response.text
            try:
                response_json = e.response.json()
            except requests.exceptions.JSONDecodeError:
                response_json = None

            if response_json:
                if "detail" in response_json:
                    text = response_json["detail"]

            raise ClientException(
                f"Got HTTP Error Code {e.response.status_code}: {text}"
            )

    return response


class APIClient:
    def __init__(self):
        os.makedirs(os.path.expanduser("~/.komo"), exist_ok=True)
        api_key = os.environ.get("KOMODO_API_KEY", None)
        if not api_key:
            api_key_file = os.path.expanduser("~/.komo/api-key")
            if not os.path.isfile(api_key_file):
                raise ClientException(f"{api_key_file} does not exist")

            with open(api_key_file, "r") as f:
                api_key = f.read().strip()

        self.api_key = api_key

    @classmethod
    def register(cls, email: str, password: str):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        payload = json.dumps(
            {
                "email": email,
                "password": password,
            },
        )

        _make_request_and_raise_for_status(
            "POST",
            f"{KOMODO_API_URL}/api/v1/auth/register",
            headers,
            None,
            payload,
        )

    def refresh_token(self):
        response = _make_request_and_raise_for_status(
            "POST",
            f"{KOMODO_API_URL}/api/v1/auth/jwt/api_key/login?api_key={self.api_key}",
            None,
            None,
            None,
        )
        auth = response.json()

        self._token = auth["token"]
        with open(KOMODO_JWT_TOKEN_FILE_PATH, "w") as f:
            f.write(self._token)

    @property
    def token(self):
        if os.path.exists(KOMODO_JWT_TOKEN_FILE_PATH):
            with open(KOMODO_JWT_TOKEN_FILE_PATH, "r") as f:
                token_contents = f.read()
                decoded_token = jwt.decode(
                    # when verify_signature is false, none of the other verification options are checked
                    # https://github.com/jpadilla/pyjwt/blob/master/jwt/api_jwt.py#L140
                    token_contents,
                    options={"verify_signature": False},
                )
                if decoded_token["exp"] < time.time():
                    self.refresh_token()
                else:
                    self._token = token_contents
                    set_user({"id": decoded_token.get("sub", "n/a")})
        else:
            self.refresh_token()

        return self._token

    @classmethod
    def get_api_key(cls, email: str, password: str):
        files = {
            "username": (None, email),
            "password": (None, password),
        }

        response = _make_request_and_raise_for_status(
            "POST",
            f"{KOMODO_API_URL}/api/v1/auth/jwt/login",
            None,
            files,
            None,
        )
        auth = response.json()

        token, token_type = auth["access_token"], auth["token_type"]

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{token_type} {token}",
        }

        response = _make_request_and_raise_for_status(
            "GET",
            f"{KOMODO_API_URL}/api/v1/auth/jwt/api_key",
            headers,
            None,
            None,
        )
        auth = response.json()

        api_key = auth["api_key"]
        return api_key

    def api_request(
        self,
        method: str,
        url: str,
        files: Dict = None,
        data: Dict = None,
    ) -> Union[Dict, List]:  # Not using | for > 1 return type for < Py 3.10 compat
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.token}",
        }

        try:
            response = _make_request_and_raise_for_status(
                method,
                url,
                headers,
                files,
                data,
            )
        except RetryError as e:
            raise ClientException(f"Failed to make request to {url}")

        try:
            retval = response.json()
        except json.JSONDecodeError:
            retval = response.text

        return retval

    def get_user_id(self):
        result = self.api_request(
            "GET",
            f"{KOMODO_API_URL}/api/v1/user-id",
        )

        return result["user-id"]

    def connect_aws(self, iam_role_arn):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/aws/connect?iam_role_arn={iam_role_arn}",
        )

    def connect_lambda(self, lambda_api_key):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/lambda_labs/connect?api_key={lambda_api_key}",
        )

    def get_workdir_download_url(self, workdir_upload_id):
        response = self.api_request(
            "GET",
            f"{KOMODO_API_URL}/api/v1/workdirs/download_url?upload_id={workdir_upload_id}",
        )
        url = response["url"]
        return url

    def _upload_workdir(self, workdir_path: str) -> str:
        printing.info(f"Collecting workdir {os.path.abspath(workdir_path)}...")
        if not os.path.exists(workdir_path):
            raise ClientException(f"Workdir {workdir_path} does not exist")
        path_size_bytes, path_file_count = _get_path_size_and_file_count(workdir_path)

        if path_size_bytes > MAX_WORKDIR_SIZE_BYTES:
            path_size_gigabytes = _bytes_to_gigabytes(path_size_bytes)
            raise ClientException(
                f"Workdir {workdir_path} is too large: {path_size_gigabytes} GB. Max"
                " size is 1 GB."
            )

        path_size_mega_bytes = _bytes_to_megabytes(path_size_bytes)
        printing.info(
            f"Workdir: files: {path_file_count}, size: {path_size_mega_bytes} MB"
        )

        with tempfile.TemporaryDirectory() as td:
            workdir_zipfile = os.path.join(td, "workdir")
            workdir_zipfile = shutil.make_archive(
                workdir_zipfile,
                "zip",
                workdir_path,
            )

            upload_info = self.api_request(
                "GET", f"{KOMODO_API_URL}/api/v1/workdirs/upload_info"
            )
            workdir_upload_id = upload_info["upload_id"]
            upload_url = upload_info["url"]
            fields = upload_info["fields"]

            with open(workdir_zipfile, "rb") as f:
                files = {"file": f}

                if workdir_path == ".":
                    workdir_str = "current working directory"
                else:
                    workdir_str = f"directory {workdir_path}"
                printing.info(f"Uploading {workdir_str}")
                _make_request_and_raise_for_status(
                    "POST",
                    upload_url,
                    None,
                    files,
                    fields,
                )
                printing.success(f"Uploaded {workdir_str}")

            return workdir_upload_id

    def launch_job(
        self,
        job_config: JobConfig,
        name: Optional[str] = None,
    ) -> Job:
        payload = job_config.model_dump(exclude_none=True, mode="json")
        if "workdir" in payload:
            payload.pop("workdir")

        if name:
            payload["name"] = name

        if job_config.workdir:
            workdir_upload_id = self._upload_workdir(job_config.workdir)
            payload["workdir_upload_id"] = workdir_upload_id

        payload = json.dumps(payload)
        job_dict = self.api_request(
            "POST", f"{KOMODO_API_URL}/api/v1/jobs", data=payload
        )
        job = Job.from_dict(job_dict)

        return job

    def get_job(self, job_id: str):
        job_dict = self.api_request("GET", f"{KOMODO_API_URL}/api/v1/jobs/{job_id}")
        job = Job.from_dict(job_dict)
        return job

    def get_jobs(self):
        job_dicts = self.api_request("GET", f"{KOMODO_API_URL}/api/v1/jobs")
        jobs = [Job.from_dict(d) for d in job_dicts]
        return jobs

    def _get_log_chunk(
        self,
        job_or_machine_id: str,
        node_index: str,
        next_token: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        is_machine: bool = False,
    ):
        if is_machine:
            url = f"{KOMODO_API_URL}/api/v1/machines/{job_or_machine_id}/setup_logs"
        else:
            url = f"{KOMODO_API_URL}/api/v1/jobs/{job_or_machine_id}/{node_index}/logs"
        if next_token or start_time or end_time:
            url += "?"

        args = []
        if next_token:
            args.append(f"next_token={next_token}")
        if start_time:
            args.append(f"start_time={start_time * 1000}")
        if end_time:
            args.append(f"end_time={end_time * 1000}")

        url += "&".join(args)

        logs = self.api_request("GET", url)

        return logs

    def print_job_logs(self, job_id: str, node_index: int, follow: bool):
        job = self.get_job(job_id)
        if job.status in [JobStatus.PENDING, JobStatus.INITIALIZING]:
            raise ClientException(f"Job {job_id} has not started")

        live = (
            job.status in JobStatus.executing_statuses()
            or job.status
            in [
                JobStatus.CANCELLED,
                JobStatus.CANCELLING,
                JobStatus.FINISHED,
                JobStatus.SHUTTING_DOWN,
            ]
            and (time.time() - job.updated_timestamp) < 30
        )

        # time between server queries, in seconds (so we don't overload the server)
        TIME_BETWEEN_QUERIES = 1
        # the number of seconds after a job finishes to wait before assuming no more logs are coming
        JOB_FINISH_WAIT_TIME = 30
        next_token = None
        job_finished_time = None
        last_query_time = time.time() - TIME_BETWEEN_QUERIES
        end_time = None
        if not follow:
            end_time = int(time.time())
        while True:
            time.sleep(max(TIME_BETWEEN_QUERIES - (time.time() - last_query_time), 0))
            response = self._get_log_chunk(
                job_id, node_index, next_token, end_time=end_time
            )
            last_query_time = time.time()

            for event in response["logs"]:
                message: str = event["message"]
                if message.endswith("\n"):
                    message = message[:-1]
                printing.info(message)

            if response["next_token"] == next_token:
                # we've reached the end of the currently available logs
                if not follow or not live:
                    break

                job = self.get_job(job_id)
                if job.status not in JobStatus.executing_statuses():
                    if live:
                        if job_finished_time is None:
                            job_finished_time = time.time()
                        elif (time.time() - job_finished_time) >= JOB_FINISH_WAIT_TIME:
                            break

            next_token = response["next_token"]

    def terminate_job(self, job_id: str):
        self.api_request("POST", f"{KOMODO_API_URL}/api/v1/jobs/{job_id}/cancel")

    def finish_job(self, job_id: str):
        self.api_request("POST", f"{KOMODO_API_URL}/api/v1/jobs/{job_id}/finish")

    def get_private_ssh_key(self) -> str:
        response = self.api_request("GET", f"{KOMODO_API_URL}/api/v1/ssh-key")
        ssh_key = response["ssh-key"]
        return ssh_key

    def launch_machine(
        self,
        machine_config: MachineConfig,
        name: str,
    ) -> Machine:
        payload = machine_config.model_dump(exclude_none=True)
        if "workdir" in payload:
            payload.pop("workdir")
        payload["name"] = name

        if machine_config.workdir:
            workdir_upload_id = self._upload_workdir(machine_config.workdir)
            payload["workdir_upload_id"] = workdir_upload_id

        payload = json.dumps(payload)
        machine_dict = self.api_request(
            "POST", f"{KOMODO_API_URL}/api/v1/machines", data=payload
        )
        machine = Machine.from_dict(machine_dict)

        return machine

    def get_machines(self) -> List[Machine]:
        machine_dicts = self.api_request("GET", f"{KOMODO_API_URL}/api/v1/machines")
        machines = [Machine.from_dict(d) for d in machine_dicts]
        return machines

    def get_machine(self, machine_id_or_name: str, is_name: bool = False) -> Machine:
        machine_dict = self.api_request(
            "GET",
            f"{KOMODO_API_URL}/api/v1/machines/{machine_id_or_name}?is_name={str(is_name).lower()}",
        )
        machine = Machine.from_dict(machine_dict)
        return machine

    def terminate_machine(self, machine_id: str):
        self.api_request("DELETE", f"{KOMODO_API_URL}/api/v1/machines/{machine_id}")

    def mark_job_as_running_setup(self, job_id: str):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/jobs/{job_id}/running_setup",
        )

    def mark_job_as_running(self, job_id: str):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/jobs/{job_id}/running",
        )

    def post_logs(self, task_id: str, task_type: str, logs: List[Dict[str, str]]):
        assert task_type in {"jobs", "machines", "services"}
        if task_type in {"jobs", "services"}:
            assert task_id.count("/") == 1

        if task_type == "machines":
            endpoint_suffix = "setup_logs"
        else:
            endpoint_suffix = "logs"

        payload = json.dumps(
            {
                "logs": logs,
            }
        )
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/{task_type}/{task_id}/{endpoint_suffix}",
            data=payload,
        )

    def print_machine_setup_logs(self, machine_id: str, follow: bool):
        machine = self.get_machine(machine_id, is_name=False)
        if machine.status in [MachineStatus.PENDING, MachineStatus.INITIALIZING]:
            raise Exception(f"Machine {machine_id} has not started")

        live = machine.status == MachineStatus.RUNNING_SETUP

        # time between server queries, in seconds (so we don't overload the server)
        TIME_BETWEEN_QUERIES = 1
        # the number of seconds after a machine finishes setup to wait before assuming no more logs are coming
        MACHINE_FINISH_WAIT_TIME = 30
        next_token = None
        machine_finished_time = None
        last_query_time = time.time() - TIME_BETWEEN_QUERIES
        end_time = None
        if not follow:
            end_time = int(time.time())
        while True:
            time.sleep(max(TIME_BETWEEN_QUERIES - (time.time() - last_query_time), 0))
            response = self._get_log_chunk(
                machine_id,
                0,
                next_token,
                end_time=end_time,
                is_machine=True,
            )
            last_query_time = time.time()

            for event in response["logs"]:
                message: str = event["message"]
                if message.endswith("\n"):
                    message = message[:-1]
                printing.info(message)

            if response["next_token"] == next_token or not response["next_token"]:
                # we've reached the end of the currently available logs
                if not follow or not live:
                    break

                machine = self.get_machine(machine_id, is_name=False)
                if machine.status != MachineStatus.RUNNING_SETUP:
                    if live:
                        if machine_finished_time is None:
                            machine_finished_time = time.time()
                        elif (
                            time.time() - machine_finished_time
                        ) >= MACHINE_FINISH_WAIT_TIME:
                            break

            next_token = response["next_token"]

    def mark_machine_as_running_setup(self, machine_id: str):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/machines/{machine_id}/running_setup",
        )

    def mark_machine_as_running(self, machine_id: str):
        self.api_request(
            "POST",
            f"{KOMODO_API_URL}/api/v1/machines/{machine_id}/running",
        )

    def get_services(self) -> List[Service]:
        service_dicts = self.api_request("GET", f"{KOMODO_API_URL}/api/v1/services")
        services = [Service.from_dict(d) for d in service_dicts]
        return services

    def get_service(self, service_id_or_name: str, is_name: bool = False) -> Machine:
        service_dict = self.api_request(
            "GET",
            f"{KOMODO_API_URL}/api/v1/services/{service_id_or_name}?is_name={str(is_name).lower()}",
        )
        service = Service.from_dict(service_dict)
        return service

    def launch_service(
        self,
        service_config: ServiceConfig,
        name: str,
    ) -> Service:
        payload = service_config.model_dump(exclude_none=True, mode="json")
        if "workdir" in payload:
            payload.pop("workdir")
        payload["name"] = name

        if service_config.workdir:
            workdir_upload_id = self._upload_workdir(service_config.workdir)
            payload["workdir_upload_id"] = workdir_upload_id

        payload = json.dumps(payload)
        service_dict = self.api_request(
            "POST", f"{KOMODO_API_URL}/api/v1/services", data=payload
        )
        service = Service.from_dict(service_dict)

        return service

    def terminate_service(
        self,
        service_id: str,
    ):
        self.api_request("DELETE", f"{KOMODO_API_URL}/api/v1/services/{service_id}")

    def get_service_replicas(self, service_id: str) -> List[ServiceReplica]:
        replica_dicts = self.api_request(
            "GET",
            f"{KOMODO_API_URL}/api/v1/services/{service_id}/replicas",
        )

        replicas = [ServiceReplica.from_dict(d) for d in replica_dicts]
        return replicas



import os
from typing import List, Optional

import click
import textwrap3
from tabulate import tabulate

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import list_machines


@click.command("list")
@handle_errors
def cmd_list():
    machines = list_machines()
    machines_to_print = [
        [
            machine.name,
            machine.status.value,
            machine.notebook_token is not None,
            machine.cloud.value if machine.cloud else None,
            machine.accelerators,
            textwrap3.fill(
                machine.status_message,
                width=50,
                replace_whitespace=False,
                drop_whitespace=False,
            ),
        ]
        for machine in machines
    ]

    printing.header(f"Found {len(machines)} Komodo machines\n", bold=True)
    printing.info(
        tabulate(
            machines_to_print,
            headers=[
                "Name",
                "Status",
                "Notebook",
                "Cloud",
                "Accelerators",
                "Message",
            ],
            tablefmt="simple_grid",
        ),
    )



import click

from komo.cli.service.cmd_launch import cmd_launch
from komo.cli.service.cmd_list import cmd_list
from komo.cli.service.cmd_terminate import cmd_terminate


@click.group()
def service():
    pass


service.add_command(cmd_list)
service.add_command(cmd_launch)
service.add_command(cmd_terminate)



import click

from komo.cli.agent.cmd_download_workdir import cmd_download_workdir
from komo.cli.agent.cmd_run import cmd_run
from komo.cli.agent.cmd_run_service_replica import cmd_run_service_replica
from komo.cli.agent.cmd_setup import cmd_setup
from komo.cli.agent.cmd_setup_machine import cmd_setup_machine
from komo.cli.agent.cmd_setup_service_replica import cmd_setup_service_replica


@click.group()
@click.pass_context
def agent(ctx: click.Context):
    pass


agent.add_command(cmd_run)
agent.add_command(cmd_setup)
agent.add_command(cmd_setup_machine)
agent.add_command(cmd_setup_service_replica)
agent.add_command(cmd_run_service_replica)
agent.add_command(cmd_download_workdir)



import json
import os
import subprocess
import sys
from datetime import datetime, timedelta

import click
import pkg_resources
import requests
from packaging import version

from komo import printing
from komo.types import ClientException


def handle_errors(fn):
    def inner(*args, **kwargs):
        try:
            fn(*args, **kwargs)
        except ClientException as e:
            printing.error(e.msg)
            exit(1)

    return inner


PACKAGE_NAME = "komo"
UPDATE_CHECK_FILE = os.path.expanduser("~/.komo/.package_update_check.json")
CHECK_INTERVAL_DAYS = 2


def _get_latest_version(package_name: str) -> str:
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()
        data = response.json()
        return data["info"]["version"]
    except requests.RequestException as e:
        return None


def _get_installed_version(package_name: str) -> str:
    try:
        return pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        return None


def _should_check_for_update() -> bool:
    if os.environ.get("__KOMODO_INTERNAL_AGENT__", None) == "1":
        return False
    if not os.path.exists(UPDATE_CHECK_FILE):
        return True

    with open(UPDATE_CHECK_FILE, "r") as f:
        try:
            data = json.load(f)
        except json.JSONDecodeError:
            return True
        last_check = datetime.fromisoformat(data.get("last_check"))
        if datetime.now() - last_check > timedelta(days=CHECK_INTERVAL_DAYS):
            return True
    return False


def _update_check_time():
    if not os.path.exists(os.path.dirname(UPDATE_CHECK_FILE)):
        os.makedirs(os.path.dirname(UPDATE_CHECK_FILE))
    with open(UPDATE_CHECK_FILE, "w") as f:
        json.dump({"last_check": datetime.now().isoformat()}, f)


def prompt_update(noconfirm: bool = False):
    if noconfirm or _should_check_for_update():
        installed_version = _get_installed_version(PACKAGE_NAME)
        latest_version = _get_latest_version(PACKAGE_NAME)

        if not installed_version or not latest_version:
            return

        installed_ver = version.parse(installed_version)
        latest_ver = version.parse(latest_version)

        if installed_ver < latest_ver:
            printing.warning(
                f"A new version of {PACKAGE_NAME} is available: {latest_version} (you"
                f" have {installed_version})."
            )
            if noconfirm or click.confirm("Do you want to update?"):
                try:
                    subprocess.check_call(
                        [
                            sys.executable,
                            "-m",
                            "pip",
                            "install",
                            "--upgrade",
                            PACKAGE_NAME,
                        ]
                    )
                    printing.success(
                        f"{PACKAGE_NAME} has been updated to version {latest_version}."
                    )
                except subprocess.CalledProcessError as e:
                    printing.error(f"Error during update: {e}")
            else:
                printing.info("Update skipped.")
        else:
            if noconfirm:
                printing.success(
                    "You are using the latest version of"
                    f" {PACKAGE_NAME} ({installed_version})."
                )

        _update_check_time()


def _get_path_size_and_file_count(path):
    total_size = 0
    file_count = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)
                file_count += 1
    return total_size, file_count


def _bytes_to_gigabytes(bytes_size):
    gigabytes = bytes_size / (1024**3)
    return round(gigabytes, 2)


def _bytes_to_megabytes(bytes_size):
    megabytes = bytes_size / (1024**2)
    return round(megabytes, 2)



import os
from typing import List, Optional

import click
import textwrap3
from tabulate import tabulate

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import list_jobs


@click.command("list")
@click.option("--limit", "-l", type=int, default=10)
@click.option("--skip", "-s", type=int, default=0)
@handle_errors
def cmd_list(
    limit: int,
    skip: int,
):
    jobs = list_jobs()
    jobs_to_print = [
        [
            job.id,
            job.name,
            job.status.value,
            job.cloud.value if job.cloud else "",
            job.accelerators,
            textwrap3.fill(
                job.status_message,
                width=50,
                replace_whitespace=False,
                drop_whitespace=False,
            ),
        ]
        for job in jobs
    ]

    printing.header(f"Found {len(jobs)} Komodo jobs\n", bold=True)
    printing.info(
        tabulate(
            jobs_to_print,
            headers=[
                "Job ID",
                "Name",
                "Status",
                "Cloud",
                "Accelerators",
                "Message",
            ],
            tablefmt="simple_grid",
        ),
    )



import click

from komo import printing
from komo.cli.utils import handle_errors, prompt_update
from komo.core import login


@click.command("update")
def cmd_update():
    prompt_update(noconfirm=True)



import os
import time
from typing import List, Optional

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_job, launch_job, print_job_logs
from komo.types import Cloud, JobConfig, JobStatus


@click.command("launch")
@click.option("--num-nodes", type=int, required=False, default=None)
@click.option("--gpus", type=str, default=None)
@click.option("--name", type=str, default=None)
@click.option("--cloud", "-c", type=str, default=None)
@click.option("--detach", "-d", is_flag=True)
@click.option("--run", "-r", type=str, default=None)
@click.argument("config_file", type=str)
@handle_errors
def cmd_launch(
    num_nodes: Optional[int],
    gpus: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    detach: bool,
    run: List[str],
    config_file: str,
):
    overrides = {}
    if num_nodes:
        overrides["num_nodes"] = num_nodes
    if gpus:
        overrides["resources/accelerators"] = gpus
    if cloud:
        overrides["resources/cloud"] = cloud
    if run:
        overrides["run"] = run
    job_config = JobConfig.from_yaml(config_file, **overrides)

    job = launch_job(
        job_config,
        name,
    )
    printing.success(f"Created job {job.name} (ID {job.id})")

    if detach:
        return

    last_message = None
    printing.info(
        f"Waiting for job {job.name} to start (this will take several minutes, you can"
        " safely Ctrl-C now)..."
    )
    while True:
        job = get_job(job.id)

        should_break = False
        error = False
        if job.status in [JobStatus.PENDING, JobStatus.INITIALIZING]:
            pass
        elif job.status in [
            JobStatus.RUNNING_SETUP,
            JobStatus.RUNNING,
            JobStatus.FINISHED,
            JobStatus.SHUTTING_DOWN,
        ]:
            should_break = True
        else:
            printing.error(f"Job status is {job.status.name}")
            should_break = True
            error = True

        if job.status_message and job.status_message != last_message:
            if error:
                printing.error(job.status_message)
            else:
                printing.info(job.status_message)

            last_message = job.status_message

        if should_break:
            break

        time.sleep(2)

    if job.status in [
        JobStatus.RUNNING_SETUP,
        JobStatus.RUNNING,
        JobStatus.FINISHED,
        JobStatus.SHUTTING_DOWN,
    ]:
        printing.success(f"Job successfully started")

        print_job_logs(job.id, 0, True)



import os
import shutil
import subprocess
import textwrap
import time
from typing import Dict, List, Optional

import yaml

from komo.api_client import KOMODO_JWT_TOKEN_FILE_PATH, APIClient
from komo.types import (
    ClientException,
    Cloud,
    Job,
    JobConfig,
    JobStatus,
    Machine,
    MachineConfig,
    MachineStatus,
    ReplicaStatus,
    Service,
    ServiceConfig,
    ServiceReplica,
    ServiceStatus,
    SSHInfoNotFound,
)


def login(api_key: str):
    if os.path.isfile(KOMODO_JWT_TOKEN_FILE_PATH):
        os.remove(KOMODO_JWT_TOKEN_FILE_PATH)
    komo_dir = os.path.expanduser("~/.komo")
    os.makedirs(komo_dir, exist_ok=True)
    api_key_file = os.path.join(komo_dir, "api-key")
    with open(api_key_file, "w") as f:
        f.write(api_key)


def launch_job(
    job_config: JobConfig,
    name: Optional[str] = None,
):
    api_client = APIClient()
    job = api_client.launch_job(
        job_config,
        name,
    )
    return job


def list_jobs() -> List[Job]:
    api_client = APIClient()
    jobs = api_client.get_jobs()
    return jobs


def get_job(job_id) -> Job:
    api_client = APIClient()
    job = api_client.get_job(job_id)
    return job


def print_job_logs(job_id, node_index: int = 0, follow: bool = False):
    api_client = APIClient()
    api_client.print_job_logs(job_id, node_index, follow)


def terminate_job(job_id):
    api_client = APIClient()
    api_client.terminate_job(job_id)


def _get_private_ssh_key() -> str:
    api_client = APIClient()
    ssh_key = api_client.get_private_ssh_key()
    return ssh_key


def ssh_job(job_id, node_index: int = 0):
    api_client = APIClient()

    job = api_client.get_job(job_id)

    if job.status not in [JobStatus.RUNNING, JobStatus.RUNNING_SETUP]:
        raise ClientException(f"Job {job_id} is not running")

    if node_index >= job.num_nodes:
        raise ClientException(
            f"Node index {node_index} is out of range for job {job_id} with"
            f" {job.num_nodes} node{'s' if job.num_nodes > 1 else ''}"
        )

    ssh_name = f"job-{job.id}"
    ssh_file = _update_ssh_config(ssh_name, job.ssh_info)

    ssh_node_name = ssh_name
    if node_index > 0:
        ssh_node_name = f"{ssh_node_name}-{job.ssh_info[node_index]['role']}"

    subprocess.call(
        [
            "ssh",
            "-t",
            ssh_node_name,
            f"cd ~/sky_workdir; bash --login",
        ]
    )

    os.remove(ssh_file)


def launch_machine(
    machine_config: MachineConfig,
    name: str,
) -> Machine:
    api_client = APIClient()
    machine = api_client.launch_machine(
        machine_config,
        name,
    )

    return machine


def list_machines() -> List[Machine]:
    api_client = APIClient()
    machines: List[Machine] = api_client.get_machines()

    running_machine_names = set(
        [m.name for m in machines if m.status == MachineStatus.RUNNING]
    )
    ssh_dir = os.path.expanduser("~/.komo/ssh")
    os.makedirs(ssh_dir, exist_ok=True)
    for ssh_machine_name in os.listdir(ssh_dir):
        machine_name = ssh_machine_name[len("machine-") :]
        if machine_name not in running_machine_names:
            os.remove(os.path.join(ssh_dir, ssh_machine_name))

    return machines


def terminate_machine(machine_name: str):
    api_client = APIClient()
    machine = api_client.get_machine(machine_name, is_name=True)
    api_client.terminate_machine(machine.id)


def get_machine(machine_name: str) -> Machine:
    api_client = APIClient()
    machine = api_client.get_machine(machine_name, is_name=True)
    return machine


def _get_private_key_file():
    ssh_dir = os.path.expanduser("~/.ssh")
    os.makedirs(ssh_dir, exist_ok=True)
    key_file = os.path.join(ssh_dir, "komodo-key")
    if not os.path.isfile(key_file):
        ssh_key = _get_private_ssh_key()
        with open(key_file, "w") as f:
            f.write(ssh_key)
        os.chmod(key_file, 0o600)

    return key_file


def ssh_machine(machine_name):
    api_client = APIClient()

    machine = api_client.get_machine(machine_name, True)
    if machine.status not in [MachineStatus.RUNNING, MachineStatus.RUNNING_SETUP]:
        raise ClientException(f"Machine {machine_name} is not running")

    ssh_name = f"machine-{machine.name}"
    ssh_file = _update_ssh_config(ssh_name, [machine.ssh_info])

    subprocess.call(
        [
            "ssh",
            "-t",
            ssh_name,
            f"cd ~/sky_workdir; bash --login",
        ]
    )

    os.remove(ssh_file)


def _setup_ssh_config():
    ssh_config_file = os.path.expanduser("~/.ssh/config")
    include_entry = "Include ~/.komo/ssh/*\n"

    config = ""
    if not os.path.isfile(ssh_config_file):
        os.makedirs(os.path.expanduser("~/.ssh"), exist_ok=True)
    else:
        with open(ssh_config_file, "r") as f:
            config = f.read()

    if include_entry in config:
        return
    config = include_entry + config

    with open(ssh_config_file, "w") as f:
        f.write(config)


def _update_ssh_config(name: str, ssh_info: List[dict]):
    ssh_key_file = _get_private_key_file()

    ssh_commands = []
    for node_ssh_info in ssh_info:
        role = node_ssh_info["role"]
        if role == "head":
            ssh_name = name
        else:
            ssh_name = f"{name}-{role}"

        ip_address = node_ssh_info.get("ip_address", None)
        user = node_ssh_info.get("ssh_user", None)
        port = node_ssh_info.get("ssh_port", None)
        proxy_command = ""

        if not ip_address or not user or not port:
            raise SSHInfoNotFound()

        docker_user = node_ssh_info.get("docker_user", None)
        docker_port = node_ssh_info.get("docker_port", None)

        if docker_user and docker_port:
            proxy_command = f"ProxyCommand ssh -i {ssh_key_file} -o Port={port} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=ERROR -o IdentitiesOnly=yes -o ExitOnForwardFailure=yes -o ServerAliveInterval=5 -o ServerAliveCountMax=3 -o ConnectTimeout=30s -o ForwardAgent=yes -W %h:%p {user}@{ip_address}"
            user = docker_user
            port = docker_port
            ip_address = "localhost"

        ssh_command = textwrap.dedent(
            f"""\
            Host {ssh_name}
                HostName {ip_address}
                User {user}
                IdentityFile {ssh_key_file}
                IdentitiesOnly yes
                ForwardAgent yes
                StrictHostKeyChecking no
                UserKnownHostsFile=/dev/null
                GlobalKnownHostsFile=/dev/null
                Port {port}
                {proxy_command}
        """.rstrip()
        )
        ssh_commands.append(ssh_command)

    ssh_file = os.path.expanduser(f"~/.komo/ssh/{name}")
    os.makedirs(os.path.expanduser("~/.komo/ssh"), exist_ok=True)
    with open(ssh_file, "w") as f:
        f.write("\n\n".join(ssh_commands))

    return ssh_file


def open_machine_in_vscode(machine_name):
    _setup_ssh_config()
    api_client = APIClient()

    code = shutil.which("code")
    if code is None:
        raise ClientException(
            "Please install the VSCode CLI"
            " (https://code.visualstudio.com/docs/editor/command-line)"
        )
    machine = api_client.get_machine(machine_name, is_name=True)
    if machine.status not in [MachineStatus.RUNNING, MachineStatus.RUNNING_SETUP]:
        raise ClientException(f"Machine {machine_name} is not running")

    ssh_name = f"machine-{machine.name}"
    _update_ssh_config(ssh_name, [machine.ssh_info])
    user = machine.ssh_info.get("docker_user", None) or machine.ssh_info["ssh_user"]
    if machine.ssh_info.get("docker_user", None) == "root":
        user = "/root"
    else:
        user = f"/home/{user}"

    subprocess.call(
        [
            "code",
            "--remote",
            f"ssh-remote+{ssh_name}",
            f"{user}/sky_workdir",
        ]
    )


def print_machine_setup_logs(machine_name: str, follow: bool):
    api_client = APIClient()
    machine = api_client.get_machine(machine_name, is_name=True)
    api_client.print_machine_setup_logs(machine.id, follow)


def list_services() -> List[Service]:
    api_client = APIClient()
    services = api_client.get_services()
    return services


def get_service(service_name: str) -> Service:
    api_client = APIClient()
    service = api_client.get_service(service_name, is_name=True)
    return service


def launch_service(
    service_config: ServiceConfig,
    name: str,
) -> Service:
    api_client = APIClient()
    service = api_client.launch_service(
        service_config,
        name,
    )

    return service


def terminate_service(
    service_name: str,
):
    api_client = APIClient()
    service = api_client.get_service(service_name, is_name=True)
    api_client.terminate_service(service.id)



import os
import time
from typing import Optional

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_service, launch_service
from komo.types import Cloud, ServiceConfig, ServiceStatus


@click.command("launch")
@click.option("--gpus", type=str, default=None)
@click.option("--cloud", "-c", type=str, default=None)
@click.option("--name", type=str, required=True)
@click.option("--detach", "-d", is_flag=True, default=False)
@click.argument("config_file", nargs=1)
@handle_errors
def cmd_launch(
    gpus: Optional[str],
    cloud: Optional[str],
    name: str,
    detach: bool,
    config_file: str,
):
    if not os.path.isfile(config_file):
        printing.error(f"{config_file} does not exist")
        exit(1)

    overrides = {}
    if gpus:
        overrides["resources/accelerators"] = gpus
    if cloud:
        overrides["resources/cloud"] = cloud

    service_config = ServiceConfig.from_yaml(config_file, **overrides)

    service = launch_service(
        service_config,
        name,
    )

    printing.success(f"Succesfully launched service {service.name}")

    if detach:
        return

    printing.info(
        f"Waiting for service {service.name} to start (this will take several minutes,"
        " you can safely Ctrl-C now)..."
    )

    if service.service.replica_policy.min_replicas == 0:
        statuses_to_wait = [
            ServiceStatus.CONTROLLER_INIT,
        ]
    else:
        statuses_to_wait = [
            ServiceStatus.CONTROLLER_INIT,
            ServiceStatus.REPLICA_INIT,
            ServiceStatus.NO_REPLICA,
        ]

    while True:
        service = get_service(service.name)

        if service.status in statuses_to_wait:
            time.sleep(5)
            continue

        break

    if service.status == ServiceStatus.READY:
        printing.success(
            f"Service {service.name} successfully started at {service.url}"
        )
    elif (
        service.service.replica_policy.min_replicas == 0
        and service.status == ServiceStatus.NO_REPLICA
    ):
        printing.success(
            f"Service {service.name} successfully started at {service.url}. There are"
            " currently no ready replicas yet, you will need to hit the endpoint to"
            " trigger a scale up."
        )
    elif service.status == ServiceStatus.SHUTTING_DOWN:
        printing.info(f"Service {service.name} was terminated")
    else:
        if service.status_message:
            printing.error(
                f"Service {service.name} failed with the following"
                f" message:\n{service.status_message}"
            )
        else:
            printing.error(
                f"Service {service.name} failed to start and currently has status"
                f" '{service.status.value}'"
            )

        exit(1)



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import terminate_machine


@click.command("terminate")
@click.argument(
    "machine_name",
    type=str,
)
@handle_errors
def cmd_terminate(machine_name: str):
    terminate_machine(machine_name)
    printing.success(f"Machine {machine_name} is being terminated")



import click

from komo.cli.job.cmd_launch import cmd_launch
from komo.cli.job.cmd_list import cmd_list
from komo.cli.job.cmd_logs import cmd_logs
from komo.cli.job.cmd_ssh import cmd_ssh
from komo.cli.job.cmd_terminate import cmd_terminate


@click.group
def job():
    pass


job.add_command(cmd_launch)
job.add_command(cmd_terminate)
job.add_command(cmd_list)
job.add_command(cmd_logs)
job.add_command(cmd_ssh)



import os
import click

from komo import printing
from komo.api_client import APIClient
from komo.cli.agent.agent import agent
from komo.cli.aws.aws import aws
from komo.cli.cmd_login import cmd_login
from komo.cli.cmd_update import cmd_update
from komo.cli.job.job import job
from komo.cli.lambda_labs.lambda_labs import lambda_labs
from komo.cli.machine.machine import machine
from komo.cli.service.service import service
from komo.cli.utils import prompt_update
from komo import printing
from komo.version import __version__
import sentry_sdk
from sentry_sdk import set_user

sentry_sdk.init(
    dsn="https://9ad4d8eed531ed220436cd5753f7bc76@o4507336556412928.ingest.us.sentry.io/4507708558999552",
    sample_rate=1.0,
    environment="production",
    enable_tracing=False,
    release=__version__,
)


@click.group()
@click.pass_context
def cli(ctx: click.Context):
    ctx.ensure_object(dict)
    # Skip update prompt if the command is 'update'
    if ctx.invoked_subcommand != "update":
        try:
            prompt_update()
        except click.exceptions.Abort:
            printing.info("\nUpdate skipped.")


cli.add_command(cmd_login)
cli.add_command(aws)
cli.add_command(lambda_labs)
cli.add_command(machine)
cli.add_command(job)
cli.add_command(service)
cli.add_command(cmd_update)

# agent is not to be used by the user, but only but running komodo workflows
if os.environ.get("__KOMODO_INTERNAL_AGENT__", None):
    cli.add_command(agent)



import json

import boto3
import botocore

from komo import printing
from komo.api_client import APIClient
from komo.types import ClientException


def connect(api_key: str):
    api_client = APIClient()
    api_client.connect_lambda(api_key)



import os
from typing import List, Optional

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import login


@click.command("login")
@handle_errors
def cmd_login():
    api_key = click.prompt("API Key")

    login(api_key)
    printing.success("You are now logged in!")



import time

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_job, print_job_logs
from komo.types import JobStatus


@click.command("logs")
@click.option("--node-index", "-i", type=int, default=0)
@click.option("--follow", "-f", is_flag=True, default=False)
@click.argument(
    "job_id",
    type=str,
)
@handle_errors
def cmd_logs(
    node_index: int,
    follow: bool,
    job_id: str,
):
    print_job_logs(job_id, node_index, follow)



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import terminate_job


@click.command("terminate")
@click.argument(
    "job_id",
    type=str,
)
@handle_errors
def cmd_terminate(job_id: str):
    terminate_job(job_id)
    printing.success(f"Job {job_id} is being terminated")



import time
from typing import Optional

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_machine, launch_machine, print_machine_setup_logs
from komo.types import Cloud, MachineConfig, MachineStatus


@click.command("launch")
@click.option("--name", type=str, required=True)
@click.option("--gpus", type=str, default=None)
@click.option("--cloud", "-c", type=str, default=None)
@click.option("--detach", "-d", is_flag=True, default=False)
@click.option("--notebook", is_flag=True, default=False)
@click.argument("config_file", nargs=1)
@handle_errors
def cmd_launch(
    name: str,
    gpus: Optional[str],
    cloud: Optional[str],
    detach: bool,
    notebook: bool,
    config_file: str,
):
    overrides = {}
    if gpus:
        overrides["resources/accelerators"] = gpus
    if cloud:
        overrides["resources/cloud"] = cloud
    if notebook:
        overrides["notebook"] = True
    machine_config = MachineConfig.from_yaml(config_file, **overrides)

    machine = launch_machine(machine_config, name)
    printing.success(f"Machine {machine.name} successfully created")

    if detach:
        return

    printing.info(
        f"Waiting for machine {machine.name} to start (this will take several minutes,"
        " you can safely Ctrl-C now)..."
    )

    last_messsage = None
    while True:
        machine = get_machine(name)

        should_break = False
        error = False
        if machine.status in [MachineStatus.PENDING, MachineStatus.INITIALIZING]:
            pass
        elif machine.status in [MachineStatus.RUNNING_SETUP, MachineStatus.RUNNING]:
            should_break = True
        else:
            should_break = True
            error = True

        if machine.status_message and machine.status_message != last_messsage:
            if error:
                printing.error(machine.status_message)
            else:
                printing.info(machine.status_message)

            last_messsage = machine.status_message

        if should_break:
            break

        time.sleep(5)

    print_machine_setup_logs(machine.name, True)

    machine = get_machine(name)
    while machine.status == MachineStatus.INITIALIZING:
        time.sleep(5)
        machine = get_machine(name)

    if machine.status == MachineStatus.RUNNING:
        printing.success(f"Machine {name} successfully created")

        if machine.notebook_url:
            printing.info(
                f"Open this link to access the notebook: {machine.notebook_url}"
            )
    else:
        printing.error(
            f"Machine {name} has status {machine.status.value} with the following"
            f" message:\n{machine.status_message}"
        )



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_machine, open_machine_in_vscode
from komo.types import MachineStatus


@click.command("vscode")
@click.argument(
    "machine_name",
    type=str,
)
@handle_errors
def cmd_vscode(machine_name: str):
    machine = get_machine(machine_name)
    if machine.status != MachineStatus.RUNNING:
        printing.error(f"Machine {machine_name} is not running")
        exit(1)

    open_machine_in_vscode(machine_name)



import os
from typing import List, Optional

import click
import textwrap3
from tabulate import tabulate

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import list_services


@click.command("list")
@handle_errors
def cmd_list():
    services = list_services()
    services_to_print = [
        [
            service.name,
            service.status.value,
            service.url,
            textwrap3.fill(
                service.status_message,
                width=50,
                replace_whitespace=False,
                drop_whitespace=False,
            ),
        ]
        for service in services
    ]

    printing.header(f"Found {len(services)} Komodo services\n", bold=True)
    printing.info(
        tabulate(
            services_to_print,
            headers=[
                "Name",
                "Status",
                "URL",
                "Message",
            ],
            tablefmt="simple_grid",
        ),
    )



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_machine, ssh_machine
from komo.types import MachineStatus


@click.command("ssh")
@click.argument(
    "machine_name",
    type=str,
)
@handle_errors
def cmd_ssh(machine_name: str):
    ssh_machine(machine_name)



import json

import boto3
import botocore

from komo import printing
from komo.api_client import APIClient
from komo.types import ClientException


def connect():
    api_client = APIClient()
    user_id = api_client.get_user_id()

    session = boto3.Session()
    iam_client = session.client("iam")
    sts_client = session.client("sts")
    try:
        aws_account_id = sts_client.get_caller_identity()["Account"]
    except botocore.exceptions.BotoCoreError as e:
        raise ClientException(
            "Unable to get AWS caller identity. Please ensure that you are logged in"
            " with AWS"
        )

    iam_role_name = f"komodo-iam-role-{user_id}"
    policy_name = f"komodo-iam-role-policy-{user_id}"
    policy_arn = f"arn:aws:iam::{aws_account_id}:policy/{policy_name}"

    try:
        try:
            role_response = iam_client.get_role(RoleName=iam_role_name)
        except iam_client.exceptions.NoSuchEntityException:
            printing.info(
                "Creating IAM Role to allow the Komodo server access to your account"
            )

            try:
                policy_response = iam_client.get_policy(PolicyArn=policy_arn)
            except iam_client.exceptions.NoSuchEntityException:
                policy_response = iam_client.create_policy(
                    PolicyName=policy_name,
                    PolicyDocument=json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": "ec2:RunInstances",
                                    "Resource": "arn:aws:ec2:*::image/ami-*",
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": "ec2:RunInstances",
                                    "Resource": [
                                        f"arn:aws:ec2:*:{aws_account_id}:instance/*",
                                        f"arn:aws:ec2:*:{aws_account_id}:network-interface/*",
                                        f"arn:aws:ec2:*:{aws_account_id}:subnet/*",
                                        f"arn:aws:ec2:*:{aws_account_id}:volume/*",
                                        f"arn:aws:ec2:*:{aws_account_id}:security-group/*",
                                    ],
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "ec2:TerminateInstances",
                                        "ec2:DeleteTags",
                                        "ec2:StartInstances",
                                        "ec2:CreateTags",
                                        "ec2:StopInstances",
                                    ],
                                    "Resource": (
                                        f"arn:aws:ec2:*:{aws_account_id}:instance/*"
                                    ),
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": ["ec2:Describe*"],
                                    "Resource": "*",
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "ec2:CreateSecurityGroup",
                                        "ec2:AuthorizeSecurityGroupIngress",
                                    ],
                                    "Resource": f"arn:aws:ec2:*:{aws_account_id}:*",
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "iam:GetRole",
                                        "iam:PassRole",
                                        "iam:CreateRole",
                                        "iam:AttachRolePolicy",
                                    ],
                                    "Resource": [
                                        f"arn:aws:iam::{aws_account_id}:role/skypilot-v1"
                                    ],
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "iam:GetInstanceProfile",
                                        "iam:CreateInstanceProfile",
                                        "iam:AddRoleToInstanceProfile",
                                    ],
                                    "Resource": f"arn:aws:iam::{aws_account_id}:instance-profile/skypilot-v1",
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": "iam:CreateServiceLinkedRole",
                                    "Resource": "*",
                                    "Condition": {
                                        "StringEquals": {
                                            "iam:AWSServiceName": "spot.amazonaws.com"
                                        }
                                    },
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": [
                                        "ec2:DeleteSecurityGroup",
                                        "ec2:ModifyInstanceAttribute",
                                    ],
                                    "Resource": f"arn:aws:ec2:*:{aws_account_id}:*",
                                },
                                {
                                    "Effect": "Allow",
                                    "Action": ["s3:*", "s3-object-lambda:*"],
                                    "Resource": "*",
                                },
                            ],
                        }
                    ),
                    Description=(
                        "These are the permissions that the Komodo server will have to"
                        " be able to manage your infrastructure"
                    ),
                )

            role_response = iam_client.create_role(
                RoleName=iam_role_name,
                AssumeRolePolicyDocument=json.dumps(
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Sid": "Statement1",
                                "Effect": "Allow",
                                "Principal": {"AWS": "211125406737"},
                                "Condition": {
                                    "StringEquals": {
                                        "sts:ExternalId": user_id,
                                    }
                                },
                                "Action": "sts:AssumeRole",
                            }
                        ],
                    }
                ),
            )

            attach_response = iam_client.attach_role_policy(
                RoleName=iam_role_name,
                PolicyArn=policy_arn,
            )
    except botocore.exceptions.BotoCoreError as e:
        raise ClientException(str(e))

    role_arn = role_response["Role"]["Arn"]
    api_client.connect_aws(role_arn)



import click

from komo.cli.lambda_labs.cmd_connect import cmd_connect


@click.group("lambda")
@click.pass_context
def lambda_labs(ctx: click.Context):
    pass


lambda_labs.add_command(cmd_connect)



import os
import subprocess
import tempfile
import time
import traceback
import zipfile

import requests
from retry.api import retry_call

from komo.api_client import APIClient


class Logger:
    _TIME_BETWEEN_LOGS = 1

    def __init__(
        self,
        api_client: APIClient,
        task_id: str,
        task_type: str,
    ):
        self.api_client = api_client
        self.task_id = task_id
        self.task_type = task_type
        self.buffer = []
        self.last_log_time = None

    def flush(self):
        if len(self.buffer) == 0:
            return

        try:
            self.api_client.post_logs(
                self.task_id,
                self.task_type,
                self.buffer,
            )
        except Exception as e:
            print(e)
            traceback.print_exc()

        self.buffer = []
        self.last_log_time = time.time()

    def flush_if_necessary(self):
        curr_time = time.time()

        if (
            not self.last_log_time
            or (curr_time - self.last_log_time) > self._TIME_BETWEEN_LOGS
        ):
            self.flush()

    def log(self, message: str):
        self.flush_if_necessary()

        self.buffer.append(
            {
                "timestamp": int(time.time() * 1000),
                "message": message,
            }
        )

    def __del__(self):
        self.flush()


def _get_setup_node_index() -> int:
    node_index = os.environ.get("SKYPILOT_SETUP_NODE_RANK", 0)
    node_index = int(node_index)
    return node_index


def _get_node_index() -> int:
    node_index = os.environ.get("SKYPILOT_NODE_RANK", 0)
    node_index = int(node_index)
    return node_index


def _execute(
    api_client: APIClient,
    task_id: str,
    task_type: str,
    script: str,
    raise_on_error: bool = True,
):
    logger = Logger(api_client, task_id, task_type)

    with tempfile.TemporaryDirectory() as td:
        script_file = os.path.join(td, "script.sh")

        with open(script_file, "w") as f:
            f.write(script)

        proc = subprocess.Popen(
            ["bash", script_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        for line in proc.stdout:
            try:
                logger.log(line.decode("utf-8"))
            except:
                pass

        proc.communicate()
        if raise_on_error and proc.returncode != 0:
            raise Exception(f"Process exited with return code {proc.returncode}")


LAMBDA_LABS_CREDENTIALS_PATH = os.path.expanduser("~/.lambda_cloud/lambda_keys")
AWS_CREDENTIALS_PATH = os.path.expanduser("~/.aws/credentials")


def _cleanup_komodo_credentials():
    # Delete cloud credentials from the environment
    if os.path.exists(LAMBDA_LABS_CREDENTIALS_PATH):
        os.remove(LAMBDA_LABS_CREDENTIALS_PATH)


def __run_command(command):
    """Run a shell command and return the output."""
    result = subprocess.run(command, shell=True, text=True, capture_output=True)
    if result.returncode != 0:
        raise Exception(f"Command failed: {command}\n{result.stderr}")
    return result.stdout


def _setup_ufw(allowed_ports):
    try:
        # Reset ufw to default settings
        __run_command("sudo ufw --force reset")
        # Default policies: deny all incoming, allow all outgoing
        __run_command("sudo ufw default deny incoming")
        __run_command("sudo ufw default allow outgoing")
        # Always allow OpenSSH (port 22)
        __run_command("sudo ufw allow OpenSSH")
        # Allow the ports provided by Cloud Runner
        for port in allowed_ports:
            __run_command(f"sudo ufw allow {port}")

        # Enable UFW
        __run_command("sudo ufw --force enable")
        print("UFW setup complete.")
    except Exception as e:
        print(f"an error occurred while setting up UFW: {e}")
        raise e


def _setup_komodo_cloud():
    """
    Cleanup credentials and setup UFW for Komodo Cloud.
    """
    try:
        is_komodo_cloud = os.environ.get("KOMODO_CLOUD", None) == "1"
        if is_komodo_cloud:
            _cleanup_komodo_credentials()
            komodo_user_requested_ports_str = os.environ.get(
                "KOMODO_USER_REQUESTED_PORTS", "[]"
            )
            if (
                komodo_user_requested_ports_str == "[]"
                or komodo_user_requested_ports_str == ""
            ):
                komodo_user_requested_ports = []
            else:
                komodo_user_requested_ports = komodo_user_requested_ports_str.split(",")
            _setup_ufw(komodo_user_requested_ports)
    except Exception as e:
        raise Exception(f"An error occurred while setting up Komodo Cloud: {e}")


def setup(job_id: str, setup_script: str):
    api_client = APIClient()
    node_index = _get_setup_node_index()
    try:
        _setup_komodo_cloud()
        api_client.mark_job_as_running_setup(job_id)
        task_id = f"{job_id}/{node_index}"
        _execute(api_client, task_id, "jobs", setup_script)
    except Exception as e:
        if node_index == 0:
            retry_call(api_client.finish_job, fargs=[job_id], tries=-1, delay=30)
        raise e


def run(job_id: str, run_script: str):
    api_client = APIClient()
    node_index = _get_node_index()
    try:
        api_client.mark_job_as_running(job_id)
        task_id = f"{job_id}/{node_index}"
        _execute(api_client, task_id, "jobs", run_script)
    finally:
        if node_index == 0:
            retry_call(api_client.finish_job, fargs=[job_id], tries=-1, delay=30)


def setup_machine(machine_id: str, setup_script: str):
    api_client = APIClient()

    # Raise an error if Komodo Cloud setup fails since if it fails we may be leaking credentials or leaving
    # all ports open.
    try:
        _setup_komodo_cloud()
    except Exception as e:
        retry_call(api_client.terminate_machine, fargs=[machine_id], tries=-1, delay=30)
        raise e

    # We don't raise on error because we still want the machine to stay running even if the
    # setup script fails (so the user can debug)
    # TODO: instead of using raise_on_error=False, catch an exception, and mark the machine
    # as "setup failed" (will have to create this endpoint on the server)
    try:
        api_client.mark_machine_as_running_setup(machine_id)
        _execute(api_client, machine_id, "machines", setup_script, raise_on_error=False)
    finally:
        retry_call(
            api_client.mark_machine_as_running, fargs=[machine_id], tries=-1, delay=30
        )


def setup_service_replica(service_id: str, replica_id: int, setup_script: str):
    _setup_komodo_cloud()
    api_client = APIClient()
    task_id = f"{service_id}/{replica_id}"
    _execute(api_client, task_id, "services", setup_script)


def run_service_replica(service_id: str, replica_id: int, run_script: str):
    api_client = APIClient()
    task_id = f"{service_id}/{replica_id}"
    _execute(api_client, task_id, "services", run_script)


def download_workdir(workdir_upload_id: str, destination: str):
    api_client = APIClient()
    os.makedirs(destination, exist_ok=True)

    download_url = api_client.get_workdir_download_url(workdir_upload_id)
    response = retry_call(
        requests.get,
        fargs=[download_url],
        tries=10,
        delay=3,
        backoff=1.2,
    )
    response.raise_for_status()
    tf = tempfile.mktemp()
    with open(tf, "wb") as f:
        f.write(response.content)

    with zipfile.ZipFile(tf, "r") as zf:
        zf.extractall(destination)

    os.remove(tf)



import click

from komo.agent.core import setup_machine


@click.command("setup-machine")
@click.option("--machine-id", type=str, required=True)
@click.option("--script", type=str, required=True)
def cmd_setup_machine(
    machine_id: str,
    script: str,
):
    setup_machine(machine_id, script)



import click


def info(msg: str, bold=False):
    click.secho(msg, fg="white", bold=bold)


def error(msg: str, bold=False):
    click.secho(msg, fg="red", bold=bold)


def warning(msg: str, bold=False, nl=True):
    click.secho(msg, fg="yellow", bold=bold, nl=nl)


def success(msg: str, bold=False):
    click.secho(msg, fg="green", bold=bold)


def header(msg: str, bold=False):
    click.secho(msg, fg="magenta", bold=bold)



import click

from komo.agent.core import run


@click.command("run")
@click.option("--job-id", type=str, required=True)
@click.option("--script", type=str, required=True)
def cmd_run(
    job_id: str,
    script: str,
):
    run(job_id, script)



import time

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_machine, print_machine_setup_logs
from komo.types import JobStatus


@click.command("setup-logs")
@click.option("--follow", "-f", is_flag=True, default=False)
@click.argument(
    "machine_name",
    type=str,
)
@handle_errors
def cmd_setup_logs(
    follow: bool,
    machine_name: str,
):
    print_machine_setup_logs(machine_name, follow)



import click

from komo.cli.utils import handle_errors
from komo.lambda_labs.connect import connect


@click.command("connect")
@handle_errors
def cmd_connect():
    api_key = click.prompt("Please enter your Lambda Labs API Key")
    connect(api_key)



"""
Komodo AI CLI
"""

from setuptools import find_packages, setup
from komo.version import __version__

with open("requirements.txt", "r") as f:
    dependencies = f.readlines()

setup(
    name="komo",
    version=__version__,
    url="https://github.com/komodoai/komo",
    license="BSD",
    author="Komodo AI",
    author_email="anish@komodoai.dev",
    description="Komodo AI CLI",
    long_description=__doc__,
    packages=find_packages(exclude=["tests", "komoproj.yaml"], include=["*"]),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    python_requires=">=3.8.0",
    install_requires=dependencies,
    entry_points={
        "console_scripts": [
            "komo = komo.cli.cli:cli",
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        "Development Status :: 4 - Beta",
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import click

from komo.agent.core import run_service_replica


@click.command("run-service-replica")
@click.option("--service-id", type=str, required=True)
@click.option("--replica-id", type=int, required=True)
@click.option("--script", type=str, required=True)
def cmd_run_service_replica(
    service_id: str,
    replica_id: int,
    script: str,
):
    run_service_replica(service_id, replica_id, script)



import webbrowser

import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import get_machine


@click.command("notebook")
@click.argument(
    "machine_name",
    type=str,
)
@handle_errors
def cmd_notebook(machine_name: str):
    machine = get_machine(machine_name)
    url = machine.notebook_url
    if not url:
        printing.error("There is no notebook running on this machine")
        exit(1)
    printing.success(f"Opening notebook at {url}")
    webbrowser.open(url)



import click

from komo import printing
from komo.cli.utils import handle_errors
from komo.core import terminate_service


@click.command("terminate")
@click.argument(
    "service_name",
    type=str,
)
@handle_errors
def cmd_terminate(service_name: str):
    terminate_service(service_name)
    printing.success(f"Service {service_name} is being terminated")



import click

from komo.cli.machine.cmd_launch import cmd_launch
from komo.cli.machine.cmd_list import cmd_list
from komo.cli.machine.cmd_notebook import cmd_notebook
from komo.cli.machine.cmd_setup_logs import cmd_setup_logs
from komo.cli.machine.cmd_ssh import cmd_ssh
from komo.cli.machine.cmd_terminate import cmd_terminate
from komo.cli.machine.cmd_vscode import cmd_vscode


@click.group()
def machine():
    pass


machine.add_command(cmd_launch)
machine.add_command(cmd_list)
machine.add_command(cmd_terminate)
machine.add_command(cmd_ssh)
machine.add_command(cmd_vscode)
machine.add_command(cmd_setup_logs)
machine.add_command(cmd_notebook)
